{# TODO: wrap all these in their own divs for easier styling #}

{% if theme_donate_url or theme_opencollective  %}
<h3 class="donation">Donate</h3>
{% endif %}

If you find marshmallow useful, please consider supporting the team with a donation:

{% if theme_donate_url %}
<p>
<a class="badge" href="{{ theme_donate_url }}">
<img src="https://img.shields.io/badge/donate-%E2%9D%A4%C2%A0-ff69b4.svg?style=flat" alt="Donate">
</a>
</p>
{% endif %}

{% if theme_opencollective %}
<p>
<a class="badge" href="https://opencollective.com/{{ theme_opencollective }}/donate" target="_blank">
  <img src="https://opencollective.com/{{ theme_opencollective }}/donate/button.png?color={{ theme_opencollective_button_color }}" width=300 />
</a>
</p>
{% endif %}

Your donation keeps marshmallow healthy and maintained.

{% if theme_tidelift_url %}
<h3 class="donation">Professional support</h3>

<p>
Professionally-supported {{ project }} is available with the
<a href="{{ theme_tidelift_url }}">Tidelift Subscription</a>.
</p>
{% endif %}
